#include "Robot.h"
#include "SistemaVisio.h"
#include "Coordenades.h"

//POSICIONS DEFINIDES AL MANIPULADOR
//230: 	Posici de recollida de la primera pea de delimitaci, situada en el palet.
//232: 	Posici de recollida de la segona pea de delimitaci, situada en el palet.
//234: 	Posici despera de larribada de la safata.
//235: 	Posici de collocaci de la primera pea de delimitaci en el punt inicial
//		de lrea de recollida (o posici de referncia).
//237: 	Posici de collocaci de la segona pea de delimitaci en el punt final de
//		lrea de recollida.
//239: 	Posici sobre la caixa on es dipositen les peces.


int main(){

	Robot rvm1;
	SistemaVisio visio;
	Coordenades coord;


	//OBRIR PORT SERIE
	cout << endl << endl << "OPERTURA DEL PORT SERIE" << endl;
	rvm1.obrirPortSerie();


	//INICIALITZAR I CALIBRAR CAMERA
	cout << endl << endl << "INICIALITZACIO DE LA CAMERA" << endl;
	visio.inicialitzarCamera();
	cout << endl << endl << "CALIBRACIO DE LA CAMERA" << endl;
	visio.calibrarCamera();


	//ACTIVAR MOVIMENT CINTA
	cout << endl << endl << "ACTIVACIO DE LA CINTA" << endl;
	rvm1.modificarSortida(10,1);


	//INICIALITZAR ROBOT
	cout << endl << endl << "INICIALITZACIO DEL ROBOT" << endl;
	rvm1.inicialitzar();
	rvm1.enviarInstruccio("MO 234,O\r");


	//ESPERAR ARRIBADA SAFATA
	cout << endl << endl << "ESPERANT LA SAFATA" << endl;
	while(!rvm1.llegirEntrada(1)){};
	

	//COLOCAR PECES DE CALIBRACIO
	cout << endl << endl << "COL.LOCACIO DE LES PECES DE CALIBRACIO" << endl;
	rvm1.enviarInstruccio("MT 230,-60,O\r");
	rvm1.enviarInstruccio("MO 230,O\r");
	rvm1.enviarInstruccio("GC\r");
	rvm1.enviarInstruccio("MT 230,-60,C\r");
	rvm1.enviarInstruccio("MO 234,C\r");
	rvm1.enviarInstruccio("MT 235,-60,C\r");
	rvm1.enviarInstruccio("MO 235,C\r");
	rvm1.enviarInstruccio("GO\r");
	rvm1.enviarInstruccio("MT 235,-60,O\r");
	rvm1.obtenirCoordenades(1,coord);
	rvm1.enviarInstruccio("MO 234,O\r");
	rvm1.enviarInstruccio("MT 232,-60,O\r");
	rvm1.enviarInstruccio("MO 232,O\r");
	rvm1.enviarInstruccio("GC\r");
	rvm1.enviarInstruccio("MT 232,-60,C\r");
	rvm1.enviarInstruccio("MO 234,C\r");
	rvm1.enviarInstruccio("MT 237,-60,C\r");
	rvm1.enviarInstruccio("MO 237,C\r");
	rvm1.enviarInstruccio("GO\r");
	rvm1.enviarInstruccio("MT 237,-60,O\r");
	rvm1.obtenirCoordenades(2,coord);
	rvm1.enviarInstruccio("MO 234,O\r");
	rvm1.enviarInstruccio("MT 234,-1,O\r"); //Repeteixo per garantir que el robot no estorba la camera


	//DETECTAR ESPAI DE TREBALL
	cout << endl << endl << "DETECCIO DE L'ESPAI DE TREBALL" << endl;
	visio.capturarImatge(true);
	visio.deteccioBlobsAR(coord);


	//RETIRAR PECES DE CALIBRACIO
	cout << endl << endl << "RETIRADA DE LES PECES DE CALIBRACIO" << endl;
	rvm1.enviarInstruccio("MT 235,-60,O\r");
	rvm1.enviarInstruccio("MO 235,O\r");
	rvm1.enviarInstruccio("GC\r");
	rvm1.enviarInstruccio("MT 235,-60,C\r");
	rvm1.enviarInstruccio("MO 234,C\r");
	rvm1.enviarInstruccio("MT 230,-60,C\r");
	rvm1.enviarInstruccio("MO 230,C\r");
	rvm1.enviarInstruccio("GO\r");
	rvm1.enviarInstruccio("MT 230,-60,O\r");
	rvm1.enviarInstruccio("MO 234,O\r");

	rvm1.enviarInstruccio("MT 237,-60,O\r");
	rvm1.enviarInstruccio("MO 237,O\r");
	rvm1.enviarInstruccio("GC\r");
	rvm1.enviarInstruccio("MT 237,-60,C\r");
	
	rvm1.modificarSortida(9,1);
	Sleep(1000);
	rvm1.modificarSortida(9,0);

	rvm1.enviarInstruccio("MO 234,C\r");
	rvm1.enviarInstruccio("MT 232,-60,C\r");
	rvm1.enviarInstruccio("MO 232,C\r");
	rvm1.enviarInstruccio("GO\r");
	rvm1.enviarInstruccio("MT 232,-60,O\r");
	rvm1.enviarInstruccio("MO 234,O\r");
	

	cout << endl << endl << "RECOLLIDA DE LES PECES" << endl;
	for(int i=0; i<1; i++){

		//ESPERAR ARRIBADA PECA
		while(!rvm1.llegirEntrada(1)){};
	

		//TROBAR PECA
		visio.capturarImatge(true);
		visio.deteccioBlob(coord);


		//AGAFAR PECA
		rvm1.moureSobrePeca(coord);	
		rvm1.enviarInstruccio("MT 240,+60,O\r");
		rvm1.enviarInstruccio("GC\r");
		rvm1.enviarInstruccio("MO 240,C\r");


		//AVISAR CINTA QUE VAGI A BUSCAR UNA ALTRA PECA
		rvm1.modificarSortida(9,1);
		Sleep(1000);
		rvm1.modificarSortida(9,0);


		//DEIXAR PECA AGAFADA AL SEU LLOC
		rvm1.enviarInstruccio("MO 234,C\r");
		rvm1.enviarInstruccio("MO 239,C\r");
		rvm1.enviarInstruccio("GO\r");
		rvm1.enviarInstruccio("MO 234,O\r");
	}


	//DESACTIVAR CINTA
	cout << endl << endl << "DESACTIVACIO DE LA CINTA" << endl;
	rvm1.modificarSortida(10,0);


	//TANCAR PORT SERIE
	cout << endl << endl << "TANCAMENT DEL PORT SERIE" << endl;
	rvm1.tancarPortSerie();


	cout << "Execucio finalitzada. Prem ENTER per sortir";
	cin.get();

	return 1;	
}
