#include "Robot.h"


/******************************************************
 * constructor per defecte
 *
 * Inicialitza la classe 
 ******************************************************/
Robot::Robot()
{
}


/******************************************************
 * destructor
 *
 * 
 ******************************************************/
Robot::~Robot()
{
}


/******************************************************
 * obrirPortSerie
 *
 * Crea el port  i configura els parmetres de comunicaci i
 * timeout
 ******************************************************/
void Robot::obrirPortSerie()
{
	//obertura del port srie en el COM1
	cout << "Creant el port serie...";

	serie = CreateFile("\\\\.\\COM1",				//nom de l'objecte, en aquest cas el port srie
						GENERIC_READ|GENERIC_WRITE,	//tipus d'accs, lectura i escriptura
						0,							//l'objecte no es pot compartir ni obrir de nou sense tancar-lo abans
						NULL,						//sense atributs de seguretat	
						OPEN_EXISTING,				//obrir objecte existent, si no existeix retorna error
						FILE_ATTRIBUTE_NORMAL,		//l'objecte no t altres atributs
						NULL);						//sense punter a una plantilla d'atributs
	
	//comprovar que s'ha obert correctament
	if (serie == INVALID_HANDLE_VALUE)	{
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio CreateFile()." << endl << endl;	
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}

	cout << "OK" << endl;

	//preparaci d'una estructura DCB a la qual es modifiquen els atributs de comunicaci
	cout << "Configurant el atributs de comunicacio...";
	
	if(GetCommState (serie,&dcb)){		
		dcb.BaudRate = BAUDRATE;	//baud rate de 9600 bps
		dcb.Parity = PARITAT;		//bit de paritat senar
		dcb.ByteSize = BYTESIZE;	//longitut dels carcters de 7 bits
		dcb.StopBits = STOPBITS;	//2 bits de stop
	}
	else{
		//tractament d'errors
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio GetCommState()." << endl << endl;
		tancarPortSerie();	
		cout << "Prem ENTER per sortir" << endl;
		cin.get();		
		exit(1);
	}
	
	//actualitzaci de la configuracio del port srie a partir de l'estructura DCB preparada
	if(!SetCommState(serie,&dcb)){
		//tractament d'errors
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio SetCommState()." << endl << endl;
		tancarPortSerie();			
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}

	cout << "OK" << endl;
	cout << "  - Baudrate: " << BAUDRATE << endl;
	cout << "  - Paritat: " << PARITAT << endl;
	cout << "  - Bytes size: " << BYTESIZE << endl;
	cout << "  - Stop bits: " << STOPBITS << endl;

	//obtenci del parmetres de timeout actuals per modificar-los
	cout << "Configurant el timeout...";
	
	if(GetCommTimeouts(serie, &timeout)){
		timeout.ReadTotalTimeoutConstant = TIMEOUT;
	}
	else{
		//tractament d'errors
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio GetCommTimeouts()." << endl << endl;
		tancarPortSerie();		
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}

	//actualitzacio del timeout
	if(!SetCommTimeouts(serie,&timeout)){
		//tractament d'errors
		int CodiError = GetLastError();
		cerr << "error a la funcio SetCommTimeouts()." << endl << endl;
		tancarPortSerie();	
			
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}

	cout << "OK" << endl;
	cout << "  - Timeout: " << TIMEOUT << endl;
}


/******************************************************
 * tancarPortSerie
 *
 * Tanca el port de comunicaci
 ******************************************************/
void Robot::tancarPortSerie(){
	cout << "Tancant el port serie...";

	//tancament del port
	if (!CloseHandle(serie)){
		//tractament d'errors
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio CloseHandle()" << endl;		
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}

	cout << "OK" << endl;
}


/******************************************************
 * inicialitzar
 *
 * Porta el robot a l'origen i modifica la seva
 * velocitat de moviment
 ******************************************************/
void Robot::inicialitzar()
{
	cout << "Inicialitzant el robot...";
	disponible();
	WriteFile(serie,"NT\r",sizeof("NT\r"),&bytesEscrits,NULL);
	cout << "OK" << endl;

	cout << "Establint la velocitat de moviment...";
	disponible();
	WriteFile(serie,"SP 6\r",sizeof("SP 6\r"),&bytesEscrits,NULL);
	cout << "OK" << endl;
}


/******************************************************
 * disponible
 *
 * Comprova si el robot no est executant una acci per
 * tal d'evitar la prduda de missatges enviats
 ******************************************************/
void Robot::disponible()
{
	//repetir consulta fins que el robot respongui
	do{
		cout << ".";

		//sollicitar informaci al robot per veure si est disponible
		WriteFile(serie,"ID\r",sizeof("ID\r"),&bytesEscrits,NULL);
		WriteFile(serie,"DR\r",sizeof("DR\r"),&bytesEscrits,NULL);
		ReadFile(serie,lectura,50,&bytesLlegits,NULL);
	}while(bytesLlegits == 0);
	//sortida del bucle, el robot est disponible
}


/******************************************************
 * enviarInstruccio
 *
 * Envia la instrucci pasada per parmetre al robot.
 * La cadena ha d'acabar amb el carcter de retorn.
 ******************************************************/
void Robot::enviarInstruccio(char instruccio[50])
{
	int i = 0;

	//mostrar misstage que s'est enviant
	cout << "Enviant ";
	while(instruccio[i] != '\r'){
		cout << instruccio[i];
		i++;
	}
	cout << "..";
	
	//compovar disponibilitat del robot abans d'enviar res
	disponible();

	//enviar instrucci
	WriteFile(serie,instruccio,50,&bytesEscrits,NULL);
	cout << "OK" << endl;
}


/******************************************************
 * modificarSortida
 *
 * Modifica els bits de sortida del robot segons els valor
 * del parmetres. Permet enviar missatges al PLC.
 ******************************************************/
void Robot::modificarSortida(int sortida, int valor)
{
	//Desactivar bit 9
	if(sortida == 9 && valor == 0){
		cout << "Enviant OB -9...";
		disponible();
		WriteFile(serie,"OB -9\r", sizeof("OB -9\r"), &bytesEscrits, NULL);
		cout << "OK" << endl;
	}
	//Activar bit 9
	else if(sortida == 9 && valor == 1){
		cout << "Enviant OB +9...";
		disponible();
		WriteFile(serie,"OB +9\r", sizeof("OB +9\r"), &bytesEscrits, NULL);
		cout << "OK" << endl;
	}
	//Desactivar bit 10
	else if(sortida == 10 && valor == 0){
		cout << "Enviant OB -10...";
		disponible();
		WriteFile(serie,"OB -10\r", sizeof("OB -10\r"), &bytesEscrits, NULL);
		cout << "OK" << endl;
	}
	//Activar bit 10
	else if(sortida == 10 && valor == 1){
		cout << "Enviant OB +10...";
		disponible();
		WriteFile(serie,"OB +10\r", sizeof("OB +10\r"), &bytesEscrits, NULL);
		cout << "OK" << endl;
	}
	//Altra
	else{
		cout << "ERROR: sortida no valida" << endl;
	}
}


/******************************************************
 * llegirEntrada
 *
 * Llegeix els bits d'entrada del robot i retorna el
 * seu valor. Permet llegir missatges del PLC.
 ******************************************************/
int Robot::llegirEntrada(int entrada)
{
	int b = 0;
	bool senyal1 = false;
	bool senyal2 = false;
	
	cout << "Llegint entrada " << entrada << "...";	
	
	//comprova que el robot estigui disponible
	disponible();
	
	do{	
		WriteFile(serie,"ID\r",sizeof("ID\r"),&bytesEscrits,NULL);
		WriteFile(serie,"DR\r",sizeof("DR\r"),&bytesEscrits,NULL);
		b = ReadFile(serie,lectura,10,&bytesLlegits,NULL);
	}while(b != 1);
	
	cout << "OK" << endl;

	//modificar els valors dels atributs senyal1 i senyal2 segons la informaci obtinguda
	switch(lectura[3]){
		case '0': 
			senyal1 = 0;
			senyal2 = 0;
			break;
		case '1':
			senyal1 = 1;
			senyal2 = 0;
			break;	
		case '2':
			senyal1 = 0;
			senyal2 = 1;
			break;	
		case '3': 
			senyal1 = 1;
			senyal2 = 1;
			break;	 	
		default:
			break;
	}

	if(entrada == 1){
		cout << "Resposta: " << senyal1 << endl;
		return senyal1;
	}
	else if(entrada == 2){
		cout << "Resposta: " << senyal2 << endl;
		return senyal2;
	}
	else{
		cout << "ERROR: entrada no valida" << endl;
		return -1;
	}
}


/******************************************************
 * obtenirCoordneades
 *
 * Sollicita al robot els valors de les coordenades de
 * quan es troba en una posici que correspongui a un
 * punt de delimitaci de l'rea de recollida. El
 * parmetre num indica si es tracta de la posici
 * inicial de l'area (1) o la final (2).
 ******************************************************/
void Robot::obtenirCoordenades(int num, Coordenades &coord)
{	
	int index = 0;
	bool trobatX = false;
	bool trobatY = false;
	bool  negatiu = false;
	double cXrobot = 0;
	double cYrobot = 0;	
	int j = 0;
	
	//obtenir coordenades de la posici actual del robot
	cout << "Enviant WH..";
	disponible();
	WriteFile(serie,"WH\r",sizeof("WH\r"),&bytesEscrits,NULL);
	cout << "OK" << endl;
	ReadFile(serie,lectura,50,&bytesLlegits,NULL);
	
	//buscar el final del missatge i marcar el final de cadena
	j = 0;
	while(lectura[j] != '\r'){
		j++;
	}
	lectura[j+1] = '\0';	
	cout << "Resposta: " << lectura << endl;
	
	//extreu-re coordenada X
	while(!trobatX){
		switch(lectura[index]){
			//valor positiu
			case '+':
				break;
			//valor negatiu
			case '-':
				negatiu = true;
				break;
			//tots els valors de la part entera recorreguts
			case '.':
				cXrobot += ((double)lectura[index+1]-48)/10;
				if(negatiu){
					cXrobot = -cXrobot;
				}
				trobatX = true;
				break;
			//valor de la part entera
			default:
				cXrobot = cXrobot*10 + ((int)lectura[index]-48);
				break;
		}
		index++;
	}

	//extreu-re coordenada Y
	index += 2;
	negatiu = false;
	while(!trobatY){
		switch(lectura[index]){
			//valor positiu
			case '+':
				break;
			//valor negatiu)
			case '-':
				negatiu = true;
				break;
			//tots els valors de la part entera recorreguts
			case '.':
				cYrobot += ((double)lectura[index+1]-48)/10;
				if(negatiu){
					cYrobot = -cYrobot;
				}
				trobatY = true;
				break;
			//valor de la part entera
			default:
				cYrobot = cYrobot*10 + ((int)lectura[index]-48);
				break;
		}
		index++;
	}

	//si es tracta de la posici inicial de l'area...
	if(num == 1){
		//guardar les coordenades
		coord.iniXrobot = cXrobot;
		coord.iniYrobot = cYrobot;

		//guardar cadena de caracters de la poisici de referncia
		j = 0;
		while(lectura[j] != '\0'){
			posInici[j] = lectura[j];
			j++;
		}
		posInici[j] = '\0';		

		//afegir un 0 a la coord. X si no t tres xifres enteres
		if(posInici[7] != '+' && posInici[7] != '-'){
			for(int h = 49; h > 1; h--){
				posInici[h] = posInici[h-1];
			}
			posInici[1] = '0';
		}
		//afegir un 0 a la coord. X si encara no t tres xifres enteres
		if(posInici[7] != '+' && posInici[7] != '-'){
			for(int h = 49; h > 1; h--){
				posInici[h] = posInici[h-1];
			}
			posInici[1] = '0';
		}
		//afegir un 0 a la coord. Y si no t tres xifres enteres
		if(posInici[14] != '+' && posInici[14] != '-'){
			for(int h = 49; h > 8; h--){
				posInici[h] = posInici[h-1];
			}
			posInici[8] = '0';
		}
		//afegir un 0 a la coord. Y si encara no t tres xifres enteres
		if(posInici[14] != '+' && posInici[14] != '-'){
			for(int h = 49; h > 8; h--){
				posInici[h] = posInici[h-1];
			}
			posInici[8] = '0';
		}
	}

	//si es tracta de la posici final de l'area...
	else if(num == 2){
		//guardar les coordenades
		coord.fiXrobot = cXrobot;
		coord.fiYrobot = cYrobot;
	}
}


/******************************************************
 * moureSobrePeca
 *
 * Colloca al robot sobre la pea a manipular un cop
 * detectada la posici de l'objecte en coordenades del
 * robot.
 ******************************************************/
void Robot::moureSobrePeca(Coordenades coord){

	int i;
	int num;

	//preparar instruccio de recollida	
	recollirPeca[0] = 'M';
	recollirPeca[1] = 'P';
	recollirPeca[2] = ' ';
	for(i = 0; i < 47; i++){
		recollirPeca[i+3] = posInici[i];
	}

	//modificar coordenada X
	num = coord.pecaXrobot * 10;
	if(num >= 0){
		recollirPeca[3] = '+';
	}
	else{
		recollirPeca[3] = '-';
		num = abs(num);
	}
	if(num > 999){
		recollirPeca[8] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[6] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[5] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[4] = (char)(num + 48);
	}
	else if(num > 99){
		recollirPeca[8] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[6] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[5] = (char)(num + 48);
		recollirPeca[4] = '0';

	}
	else{
		recollirPeca[8] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[6] = (char)(num + 48);
		recollirPeca[5] = '0';
		recollirPeca[4] = '0';
	}

	//modificar coordenada Y
	num = coord.pecaYrobot * 10;
	if(num >= 0){
		recollirPeca[10] = '+';
	}
	else{
		recollirPeca[10] = '-';
		num = abs(num);
	}
	if(num > 999){
		recollirPeca[15] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[13] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[12] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[11] = (char)(num + 48);
	}
	else if(num > 99){
		recollirPeca[15] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[13] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[12] = (char)(num + 48);
		recollirPeca[11] = '0';

	}
	else{
		recollirPeca[15] = (char)(num % 10 + 48);
		num = num/10;
		recollirPeca[13] = (char)(num + 48);
		recollirPeca[12] = '0';
		recollirPeca[11] = '0';
	}

	//enviar instrucci al robot perque vagi a recollir la pea
	cout << "Enviant " << recollirPeca << endl;
	disponible();
	WriteFile(serie,recollirPeca,sizeof(recollirPeca),&bytesEscrits,NULL);
	cout << "OK" << endl;

	//guardar la localitzaci actual a la posici 240
	cout << "Enviant HE 240..";
	disponible();
	WriteFile(serie,"HE 240\r",sizeof("HE 240\r"),&bytesEscrits,NULL);
	cout << "OK" << endl;
}
