#include "SistemaVisio.h"


/******************************************************
 * constructor per defecte
 *
 * Inicialitza la classe
 ******************************************************/
SistemaVisio::SistemaVisio()
{
	m_iActiveBoard = 0;
	param_intrinsecs = cvCreateMat(3,3,CV_32FC1);
	coef_distorsio = cvCreateMat(1,4,CV_32FC1);
	calibrada = false;
}


/******************************************************
 * destructor
 *
 * 
 ******************************************************/
SistemaVisio::~SistemaVisio()
{
}


/******************************************************
 * inicialitzarCamera
 *
 * Inicialitza la camera i n'obte els atributs. Reserva
 * l'espai de memoria per a la captura, aixi com espai 
 * per a la captura binaritzada i per a la corretgida.
 ******************************************************/
void SistemaVisio::inicialitzarCamera()
{
	//Creaci d'un objecte de captura d'imatges
	cout << "Creant objecte de captura...";

	if (!(m_iCap = IfxCreateCaptureModule("NSP", m_iActiveBoard, "nsptest.txt"))) {
		if (!(m_iCap = IfxCreateCaptureModule("NSP", 0))) {
			//Tractament d'error
			int CodiError = GetLastError();
			cerr << "ERROR a la funcio IfxCreateCaptureModule()" << endl << endl;			
			cout << "Prem ENTER per sortir" << endl;
			cin.get();	
			exit(CodiError);
		}
		m_iCap->ProcessCameraFilesInDir("camdb", TRUE);
	}

	//Creaci de la cmera situada al port 0
	m_cam = m_iCap->GetCam(0);

	//Adquisicio dels atributs de la camera
	m_cam->GetAttr(&camAttr);
	ampladaImatge = camAttr.dwWidth;
	alcadaImatge = camAttr.dwHeight;

	cout << "OK" << endl;
	cout << "  - Amplada de la imatge: " << ampladaImatge << endl;
	cout << "  - Alcada de la imatge: " << alcadaImatge << endl;

	//Reserva d'espai de memoria per la captura i el seu tractament
	cout << "Reservant espais de captura...";
	
	if(!(m_imgBuf = (BYTE*)GlobalAllocPtr(GMEM_FIXED, ampladaImatge * alcadaImatge * camAttr.dwBytesPerPixel))){
		//Tractament d'error
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio GlobalAllocPtr() de m_imgBuf" << endl << endl;			
		cout << "Prem ENTER per sortir" << endl;
		cin.get();	
		exit(CodiError);	
	}
	if(!(m_imgBufCor = (BYTE*)GlobalAllocPtr(GMEM_FIXED, ampladaImatge * alcadaImatge * camAttr.dwBytesPerPixel))){
		//Tractament d'error
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio GlobalAllocPtr() de m_imgBufCor" << endl << endl;			
		cout << "Prem ENTER per sortir" << endl;
		cin.get();	
		exit(CodiError);	
	}
	if(!(m_imgBufCorBin = (BYTE*)GlobalAllocPtr(GMEM_FIXED, ampladaImatge * alcadaImatge * camAttr.dwBytesPerPixel))){
		//Tractament d'error
		int CodiError = GetLastError();
		cerr << "ERROR a la funcio GlobalAllocPtr() de m_imgBufCorBinf" << endl << endl;			
		cout << "Prem ENTER per sortir" << endl;
		cin.get();	
		exit(CodiError);	
	}

	cout << "OK" << endl;
}


/******************************************************
 * calibrarCamera
 *
 * Calibra la camera a traves del patro d'un tauler
 * d'escacs, obtinguent els seus parametres intrinsecs
 * i el coeficient de distorsio.
 ******************************************************/
void SistemaVisio::calibrarCamera()
{
	IplImage *img_tauler;			//imatge del tauler d'escacs
	IplImage *img_tauler_cantonades;//imatge del tauler amb les cantonades senyalades
	CvSize tamany_tauler;			//tamany del tauler segons nombre de cantonades
	CvPoint2D32f* cantonades;		//punts on es troben les cantonades
	CvMat *num_punts;
	CvMat *punts_objecte; 
	CvMat *punts_imatge;
	int cantonades_totals;
	int cantonades_trobades;
	int resultat;
	int i;


	cout << "Situa el tauler de calibracio sobre la safata i prem ENTER" ;
	cin.get();
	cout << "Calibrant la camera...";

	//capturar imatge al buffer
	m_cam->Snap(m_imgBuf);

	//escriure la imatge en un fitxer BMP
	m_imgFile = IfxCreateImgFile("C:\\tauler.bmp");
	if (m_imgFile){
		m_imgFile->WriteFile(m_imgBuf, camAttr.dwWidth, camAttr.dwHeight, camAttr.dwBitsPerPixel);
	}
	else{
		cout << "no s'ha pogut guardar la captura al disc ...";
	}

	//crear IplImatge amb el buffer de la captura
	img_tauler = cvCreateImage(cvSize(ampladaImatge,alcadaImatge), IPL_DEPTH_8U, 1);
	for(i=0; i < img_tauler->imageSize; i++){
		img_tauler->imageData[i] = m_imgBuf[i];
	}

	//trobar les cantonades d'un tauler d'escacs
	tamany_tauler = cvSize(CANTONADES_TAULER_X,CANTONADES_TAULER_Y);	
	cantonades_totals = CANTONADES_TAULER_X*CANTONADES_TAULER_Y;	
	cantonades = (CvPoint2D32f*)cvAlloc(cantonades_totals*sizeof(CvPoint2D32f));	
	cantonades_trobades = 0;
	resultat = cvFindChessboardCorners(img_tauler, tamany_tauler, cantonades, &cantonades_trobades, CV_CALIB_CB_ADAPTIVE_THRESH);
	//mirar si s'han trobat tots les cantonades (si no, resultat = 0)
	if(resultat != 0){
		//millorar la precisio de la localitzaci de les cantonades
		cvFindCornerSubPix(img_tauler, cantonades, cantonades_trobades, tamany_tauler, cvSize(-1,-1), cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,20,0.01));
	
		//dibuixar les cantonades identificades a la imatge
		img_tauler_cantonades = cvCloneImage(img_tauler);
		cvDrawChessboardCorners(img_tauler_cantonades, tamany_tauler, cantonades, cantonades_trobades, resultat);
		cvSaveImage("C:\\taulerCantonades.bmp", img_tauler_cantonades);
			
		//crear matriu de punts objecte i imatge
		num_punts = cvCreateMat(1,1,CV_32SC1);
		punts_objecte = cvCreateMat(cantonades_totals,3,CV_32FC1);
		punts_imatge = cvCreateMat(cantonades_totals,2,CV_32FC1);	
		cvSetData(punts_imatge, cantonades, sizeof(CvPoint2D32f));
		cvSetData(num_punts, &cantonades_trobades, sizeof(int));
		
		// omplir la matriu de punts objecte
		InitCorners3D(punts_objecte, tamany_tauler, 1, TAMANY_QUADRAT);

		//calcular parametres de calibracio
		cvCalibrateCamera2(punts_objecte, punts_imatge, num_punts, cvSize(ampladaImatge,alcadaImatge), param_intrinsecs, coef_distorsio, NULL, NULL, 0);
		double fx = cvmGet(param_intrinsecs, 0, 0);
		double fy = cvmGet(param_intrinsecs, 1, 1);
		double cx = cvmGet(param_intrinsecs, 0, 2);
		double cy = cvmGet(param_intrinsecs, 1, 2);
		double k1 = cvmGet(coef_distorsio, 0, 0);
		double k2 = cvmGet(coef_distorsio, 0, 1);
		double p1 = cvmGet(coef_distorsio, 0, 2);
		double p2 = cvmGet(coef_distorsio, 0, 3);
		
		calibrada = true;
	}
	else{
		cerr << " ERROR: no s'han localitzat totes les cantonades" << endl;
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}

	cout << "OK" << endl;
	cout << "Retira el tauler de calibracio i prem ENTER";
	cin.get();
}


/******************************************************
 * capturarImatge
 *
 * Realitza una captura, busca el llindar optim utilitzant
 * l'algoritme otsu, binaritza la imatge i, finalment,
 * la corregeix si s'ha demanat per parametre (requereix
 * haver calibrat la camera previament.
 ******************************************************/
void SistemaVisio::capturarImatge(bool corregir)
{
	int llindar = 0;	//Llindar optim
	int i;

	cout << "Capturant imatge...";

	//capturar imatge al buffer
	m_cam->Snap(m_imgBuf);
	cout << "OK" << endl;

	//escriure la imatge en un fitxer BMP
	cout << "Guardant captura al disc...";
	m_imgFile = IfxCreateImgFile("C:\\cap.bmp");
	if (m_imgFile){
		m_imgFile->WriteFile(m_imgBuf, camAttr.dwWidth, camAttr.dwHeight, camAttr.dwBitsPerPixel);
	}
	else{
		cout << "ERROR al guardar la captura a disc...";
	}
	cout << "OK" << endl;

	//corregir la imatge si demanat per parametre (requereix calibracio)
	if(corregir && calibrada){
		corregirImatge();
	}
	else{
		cout << "ERROR: no es pot corregir la imatge sense haver calibrat abans..." << endl;
	}
	
	//calcular el llindar optim - Otsu
	cout << "Calculant llindar optim...";
	llindar = otsu(m_imgBufCor, 1, ampladaImatge * alcadaImatge, 0, 0, ampladaImatge * alcadaImatge, 1, 0);
	cout << "OK" << endl;

	//binaritzar
	cout << "Binaritzant captura...";
	for(i=0; i<ampladaImatge * alcadaImatge; i++){		
		if((int)m_imgBufCor[i] < llindar) //llindar
			m_imgBufCorBin[i]=0;
		else
			m_imgBufCorBin[i]=255;
	}
	cout << "OK" << endl;

	//escriure la imatge en un fitxer BMP
	cout << "Guardant captura binaritzada al disc...";
	m_imgFile = IfxCreateImgFile("C:\\cap_cor_bin.bmp");
	if (m_imgFile){
		m_imgFile->WriteFile(m_imgBufCorBin, camAttr.dwWidth, camAttr.dwHeight, camAttr.dwBitsPerPixel);
	}
	else{
		cout << "ERROR al guardar la captura binaritzada a disc...";
	}
	cout << "OK" << endl;		
}


/******************************************************
 * corregirImatge
 *
 * Corregeix la imatge segons els parmetres obtinguts a travs
 * de la calibracio.
 ******************************************************/
void SistemaVisio::corregirImatge()
{
	IplImage *img_original;
	IplImage *img_corretgida;
	int i;

	cout << "Corretgint imatge...";

	//crear IplImatge amb el buffer de la captura
	img_original = cvCreateImage(cvSize(ampladaImatge,alcadaImatge), IPL_DEPTH_8U, 1);
	for(i=0; i < img_original->imageSize; i++){
		img_original->imageData[i] = m_imgBuf[i];
	}
	//corregir i guardar
	img_corretgida = cvCloneImage(img_original);
	cvUndistort2(img_original, img_corretgida, param_intrinsecs, coef_distorsio);
	cout << "Guardant captura corretgida al disc...";
	cvSaveImage("C:\\cap_cor.bmp",img_corretgida);

	//volcar IplImage de la captura corretgida a buffer	
	for(i=0; i < img_corretgida->imageSize; i++){
		m_imgBufCor[i] = img_corretgida->imageData[i];
	}

	cout << "OK" << endl;
}


/******************************************************
 * deteccioBlobsAR
 *
 * Captura la imatge i detecta la posici dels dos blobs de 
 * delimitacio per guardar les seves coordenades i calcular
 * l'area de recollida.
 ******************************************************/
void SistemaVisio::deteccioBlobsAR(Coordenades &coord)
{
	IplImage *img_blobs;	//Punter a la imatge IPL que cont els blobs
	CBlobResult blobs;		//Blobs identificats a la imatge
	int num_blobs;			//Nombre de blobs
	int i;

	cout << "Detectant blobs de calibracio...";

	//crear IplImatge amb el buffer de la captura corretgida
	img_blobs = cvCreateImage(cvSize(ampladaImatge,alcadaImatge), IPL_DEPTH_8U, 1);
	for(i=0; i < img_blobs->imageSize; i++){
		img_blobs->imageData[i] = m_imgBufCorBin[i];

	}

	//localitzar blobs
	blobs = CBlobResult(img_blobs, NULL, 50, true);
	blobs.Filter(blobs, B_INCLUDE, CBlobGetArea(), B_GREATER, 10000 );
	blobs.Filter(blobs, B_INCLUDE, CBlobGetArea(), B_LESS, 14000 );
	blobs.Filter(blobs, B_INCLUDE, CBlobGetCompactness(), B_GREATER, 1);
	blobs.Filter(blobs, B_INCLUDE, CBlobGetCompactness(), B_LESS, 1.3 );
	num_blobs = blobs.GetNumBlobs();
	double_stl_vector area = blobs.GetSTLResult(CBlobGetArea());
	double_stl_vector circ = blobs.GetSTLResult(CBlobGetCompactness());
	double_stl_vector cenX = blobs.GetSTLResult(CBlobGetXCenter());
	double_stl_vector cenY = blobs.GetSTLResult(CBlobGetYCenter());

	cout << "OK" << endl;

	//mostrar informaci dels blobs trobats
	cout << "La imatge conte " << num_blobs << " blobs" << endl;
	for(i=0; i<num_blobs; i++){
		cout << "BLOB " << i << ":" << endl;
		cout << "	Area (pixel): " << area[i] << endl;
		cout << "	Circularitat: " << circ[i] << endl;
		cout << "	Centre (x,y): "<< cenX[i] << "," << cenY[i] << endl;
	}

	//comprovar que hi ha els dos blobs de calibraci, guardar les seves coordenades i calcular l'espai de treball
	if(num_blobs == 2){		
		coord.iniXimatge = cenX[0];
		coord.iniYimatge = cenY[0];
		coord.fiXimatge = cenX[1];
		coord.fiYimatge = cenY[1];
		coord.ampladaARimatge = coord.fiXimatge - coord.iniXimatge;
		coord.alcadaARimatge = coord.fiYimatge - coord.iniYimatge;
		coord.ampladaARrobot = abs(coord.iniXrobot - coord.fiXrobot);
		coord.alcadaARrobot = abs(coord.iniYrobot - coord.fiYrobot);
	}
	//no s'han detectat els dos blobs
	else{
		cerr << "ERROR: numero de blobs incorrecte, han de ser dos" << endl;		
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}
}


/******************************************************
 * deteccioBlob
 *
 * Captura la imatge i detecta la posicio del blob a
 * agafar. Si es troba dinsl'area de recolllida guarda
 * les seves coordenades i troba les equivalents en 
 * les coordenades del robot.
 ******************************************************/
void SistemaVisio::deteccioBlob(Coordenades &coord)
{
	IplImage *img_blobs;	//Punter a la imatge IPL que cont els blobs
	CBlobResult blobs;		//Blobs identificats a la imatge
	double pecaXimatge;
	double pecaYimatge;
	int num_blobs;
	int i;

	cout << "Detectant blobs...";

	//crear IplImatge amb el buffer de la captura corretida
	img_blobs = cvCreateImage(cvSize(ampladaImatge,alcadaImatge), IPL_DEPTH_8U, 1);
	for(i=0; i < img_blobs->imageSize; i++){
		img_blobs->imageData[i] = m_imgBufCorBin[i];
	}

	//localitzar blobs
	blobs = CBlobResult(img_blobs, NULL, 50, true);
	blobs.Filter(blobs, B_INCLUDE, CBlobGetArea(), B_GREATER, 10000 );
	blobs.Filter(blobs, B_INCLUDE, CBlobGetArea(), B_LESS, 12000 );
	blobs.Filter(blobs, B_INCLUDE, CBlobGetCompactness(), B_GREATER, 1);
	blobs.Filter(blobs, B_INCLUDE, CBlobGetCompactness(), B_LESS, 1.3 );
	num_blobs = blobs.GetNumBlobs();
	double_stl_vector area = blobs.GetSTLResult(CBlobGetArea());
	double_stl_vector circ = blobs.GetSTLResult(CBlobGetCompactness());
	double_stl_vector cenX = blobs.GetSTLResult(CBlobGetXCenter());
	double_stl_vector cenY = blobs.GetSTLResult(CBlobGetYCenter());

	cout << "OK";

	//mostrar informaci dels blobs trobats
	cout << endl << "La imatge conte " << num_blobs << " blob/s" << endl;
	for(i=0; i<num_blobs; i++){
		cout << "BLOB " << i << ":" << endl;
		cout << "	Area (pixel): " << area[i] << endl;
		cout << "	Circularitat: " << circ[i] << endl;
		cout << "	Centre (x,y): "<< cenX[i] << "," << cenY[i] << endl << endl;
	}

	//hi ha un sol objecte a la safata
	if(num_blobs == 1){
		pecaXimatge = cenX[0];
		pecaYimatge = cenY[0];
	}
	//no hi ha cap objecte
	else if (num_blobs == 0){
		cerr << "ERROR: no s'ha detectat la peca..." << endl;
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}
	//hi ha ms d'un objecte
	else{
		cerr << "ERROR: hi ha ms d'un objecte o s'han capturat artefactes" << endl;
		cout << "Prem ENTER per sortir" << endl;
		cin.get();
		exit(1);
	}

	//comprovar que la peca estigui dins els lmits de l'espai de treball
	cout << "Comprovant que la peca estigui dins els limits...";
	if(pecaXimatge >= coord.iniXimatge && pecaXimatge <= coord.fiXimatge && pecaYimatge >= coord.iniYimatge && pecaYimatge <= coord.fiYimatge){
		cout << "OK" << endl;
		//conversi de coordenades imatge a robot
		coord.pecaXrobot = coord.iniXrobot - ((pecaYimatge - coord.iniYimatge) * coord.ampladaARrobot / coord.alcadaARimatge);
		coord.pecaYrobot = coord.iniYrobot - ((pecaXimatge - coord.iniXimatge) * coord.alcadaARrobot / coord.ampladaARimatge);
	}
	//peca fora dels lmits
	else{
		cerr << "ERROR: peca fora de l'espai de treball..." << endl;
		exit(1);
	}
}


/******************************************************
 * InitCorners3D
 *
 * Omple la matriu de punts d'objecte.
 ******************************************************/
void SistemaVisio::InitCorners3D(CvMat *Corners3D, CvSize ChessBoardSize, int NImages, float SquareSize)
{
	int CurrentImage = 0;
	int CurrentRow = 0;
	int CurrentColumn = 0;
	int NPoints = ChessBoardSize.height*ChessBoardSize.width;
	float * temppoints = new float[NImages*NPoints*3];

	// for now, assuming we're row-scanning
	for (CurrentImage = 0 ; CurrentImage < NImages ; CurrentImage++){
		for (CurrentRow = 0; CurrentRow < ChessBoardSize.height; CurrentRow++){
			for (CurrentColumn = 0; CurrentColumn < ChessBoardSize.width; CurrentColumn++){
				temppoints[(CurrentImage*NPoints*3)+(CurrentRow*ChessBoardSize.width + CurrentColumn)*3]=(float)CurrentRow*SquareSize;
				temppoints[(CurrentImage*NPoints*3)+(CurrentRow*ChessBoardSize.width + CurrentColumn)*3+1]=(float)CurrentColumn*SquareSize;
				temppoints[(CurrentImage*NPoints*3)+(CurrentRow*ChessBoardSize.width + CurrentColumn)*3+2]=0.f;
			}
		}
	}
	(*Corners3D) = cvMat(NImages*NPoints,3,CV_32FC1, temppoints);
}


/******************************************************
 * otsu
 *
 * Algoritme per trobar el llindar ptim per binaritzar la imatge.
 
 /*
 the following code was send by Ryan Dibble <dibbler@umich.edu>
 
  The algorithm is very simple but works good hopefully.
 
  Compare the grayscale histogram with a mass density diagram:
  I think the algorithm is a kind of
  divide a body into two parts in a way that the mass 
  centers have the largest distance from each other,
  the function is weighted in a way that same masses have a advantage
  
  - otsu algorithm is failing on diskrete multi color images
  
  TODO:
    RGB: do the same with all colors (CMYG?) seperately 

    test: hardest case = two colors
       bbg: test done, using a two color gray file. Output:
       # OTSU: thresholdValue = 43 gmin=43 gmax=188

 my changes:
   - float -> double
   - debug option added (vvv & 1..2)
   - **image => *image,  &image[i][1] => &image[i*cols+1]
   - do only count pixels near contrast regions
     this makes otsu much better for shadowed fonts or multi colored text 
     on white background

 (m) Joerg Schulenburg (see README for email address)
 ******************************************************/
int SistemaVisio::otsu (unsigned char *image, int rows, int cols, int x0, int y0, int dx, int dy, int vvv)
{
	unsigned char *np;    // pointer to position in the image we are working with
	unsigned char op1, op2;   // predecessor of pixel *np (start value)
	int maxc=0;           // maximum contrast (start value) 
	int thresholdValue=1; // value we will threshold at
	int ihist[256];       // image histogram
	int chist[256];       // contrast histogram

	int i, j, k;          // various counters
	int is, i1, i2, ns, n1, n2, gmin, gmax;
	double m1, m2, sum, csum, fmax, sb;

	// zero out histogram ...
	memset(ihist, 0, sizeof(ihist));
	memset(chist, 0, sizeof(chist));
	op1=op2=0;

	gmin=255; gmax=0; k=1;//k=dy/512+1;
	// v0.43 first get max contrast, dont do it together with next step
	//  because it failes if we have pattern as background (on top)
	for (i =  0; i <  dy ; i+=k) {
	np = &image[(y0+i)*cols+x0];
	for (j = 0; j < dx ; j++) {
	  ihist[*np]++;
	  if(*np > gmax) gmax=*np;
	  if(*np < gmin) gmin=*np;
	  if (Abs(*np-op1)>maxc) maxc=Abs(*np-op1); /* new maximum contrast */
	  if (Abs(*np-op2)>maxc) maxc=Abs(*np-op2); /* new maximum contrast */
	  /* we hope that maxc will be find its maximum very fast */
	  op2=op1;    /* shift old pixel to next older */
	  op1=*np;     /* store old pixel for contrast check */
	  np++;       /* next pixel */
	}
	}

	// generate the histogram
	// Aug06 images with large white or black homogeneous
	//   areas give bad results, so we only add pixels on contrast edges
	for (i =  0; i <  dy ; i+=k) {
	np = &image[(y0+i)*cols+x0];
	for (j = 0; j < dx ; j++) {
	  if (Abs(*np-op1)>maxc/4
	   || Abs(*np-op2)>maxc/4)
		 chist[*np]++; // count only relevant pixels
	  op2=op1;    /* shift old pixel to next older */
	  op1=*np;     /* store old pixel for contrast check */
	  np++;       /* next pixel */
	}
	}

	// set up everything
	sum = csum = 0.0;
	ns = 0;
	is = 0;

	for (k = 0; k <= 255; k++) {
	sum += (double) k * (double) chist[k];  /* x*f(x) cmass moment */
	ns  += chist[k];                        /*  f(x)    cmass      */
	is  += ihist[k];                        /*  f(x)    imass      */
	// Debug: output to out_hist.dat?
	// fprintf(stderr,"\chistogram %3d %6d (brightness weight)", k, ihist[k]);
	}

	if (!ns) {
	// if n has no value we have problems...
	fprintf (stderr, "NOT NORMAL, thresholdValue = 160\n");
	return (160);
	}

	// ToDo: only care about extremas in a 3 pixel environment
	//       check if there are more than 2 mass centers (more colors)
	//       return object colors and color radius instead of threshold value
	//        also the reagion, where colored objects are found
	//       what if more than one background color? no otsu at all?
	//       whats background? box with lot of other boxes in it
	//       threshold each box (examples/invers.png,colors.png) 
	//  get maximum white and minimum black pixel color (possible range)
	//    check range between them for low..high contrast ???
	// typical scenes (which must be covered): 
	//    - white page with text of different colors (gray values)
	//    - binear page: background (gray=1) + black text (gray=0)
	//    - text mixed with big (dark) images
	//  ToDo: recursive clustering for maximum multipol moments?
	//  idea: normalize ihist to max=1024 before otsu?

	// do the otsu global thresholding method

	if ((vvv&1)) // Debug
	 fprintf(stderr,"# OTSU: value ihist chist mass_dipol_moment\n");
	fmax = -1.0;
	n1 = 0;
	for (k = 0; k < 255; k++) {
	n1 += chist[k];          // left  mass (integration)
	if (!n1) continue;       // we need at least one foreground pixel
	n2 = ns - n1;            // right mass (num pixels - left mass)
	if (n2 == 0) break;      // we need at least one background pixel
	csum += (double) k *chist[k];  // left mass moment
	m1 =        csum  / n1;        // left  mass center (black chars)
	m2 = (sum - csum) / n2;        // right mass center (white background)
	// max. dipol moment?
	// orig: sb = (double) n1 *(double) n2 * (m1 - m2) * (m1 - m2);
	sb = (double) n1 *(double) n2 * (m2 - m1); // seems to be better Aug06
	/* bbg: note: can be optimized. */
	if (sb > fmax) {
	  fmax = sb;
	  thresholdValue = k + 1;
	  // thresholdValue = (m1 + 3 * m2) / 4;
	}
	if ((vvv&1) && ihist[k]) // Debug
	 fprintf(stderr,"# OTSU: %3d %6d %6d %8.2f\n",
	   k, ihist[k], chist[k],
	   sb/(dx*dy));  /* normalized dipol moment */
	}
	// ToDo: error = left/right point where sb is 90% of maximum?
	// now we count all pixels for background detection
	i1 = 0;
	for (k = 0; k < thresholdValue; k++) {
	i1 += ihist[k];          // left  mass (integration)
	}
	i2 = is - i1;            // right mass (num pixels - left mass)

	// at this point we have our thresholding value
	// black_char: value<cs,  white_background: value>=cs

	// can it happen? check for sureness
	if (thresholdValue >  gmax) {
	fprintf(stderr,"# OTSU: thresholdValue >gmax\n");
	thresholdValue = gmax;
	}
	if (thresholdValue <= gmin) {
	fprintf(stderr,"# OTSU: thresholdValue<=gmin\n");
	thresholdValue = gmin+1;
	}

	// debug code to display thresholding values
	if ( vvv & 1 )
	fprintf(stderr,"# OTSU: thresholdValue = %d gmin=%d gmax=%d cmax=%d"
				 " i= %d %d\n",
	 thresholdValue, gmin, gmax, maxc, i1, i2);

	if (i1>=4*i2) { // black>=4*white, obviously black is background
	if ( vvv & 1 )
	  fprintf(stderr,"# OTSU: invert the image\n");
	// we do inversion here (no data lost)
	for (i =  0; i <  dy ; i++) {
	  np = &image[(y0+i)*cols+x0];
	  for (j = 0; j < dx ; j++) {
		*np=255-*np;
		np++;       /* next pixel */
	  }
	}
	thresholdValue=255-thresholdValue+1;
	}

	return(thresholdValue); 
	/* range: 0 < thresholdValue <= 255, example: 1 on b/w images */
	/* 0..threshold-1 is foreground */
	/* threshold..255 is background */
	/* ToDo:  min=blackmasscenter/2,thresh,max=(whitemasscenter+255)/2 */
}
