#include "Preparacio.h"


using namespace std;


int main(){

	//Creacio d'un objecte de captura d'imatges
	if (!(m_iCap = IfxCreateCaptureModule("NSP", m_iActiveBoard, "nsptest.txt"))) {
		if (!(m_iCap = IfxCreateCaptureModule("NSP", 0))) {
			//Tractament d'error
			int CodiError = GetLastError();
			cout << "ERROR a la funcio IfxCreateCaptureModule()" << endl << endl;			
			cout << "Prem ENTER per sortir" << endl;
			cin.get();	
			exit(CodiError);
		}
		//Si el fitxer de configuracio no hi es, s'afegeixen totes les deficions
		//de la base de dades del seguent directori
		m_iCap->ProcessCameraFilesInDir("camdb", TRUE);
	}


	//Creacio de l'objecte corresponent a la camera del port 0
	m_cam = m_iCap->GetCam(0);

	
	//Adquisicio dels atributs de la camera
	m_cam->GetAttr(&camAttr);


	//Reserva d'espai de memoria per la captura i el seu tractament
	double tamany = camAttr.dwWidth * camAttr.dwHeight;
	m_imgBuf = (BYTE*)GlobalAllocPtr(GMEM_FIXED, tamany * camAttr.dwBytesPerPixel);
	m_imgBufBin = (BYTE*)GlobalAllocPtr(GMEM_FIXED, tamany * camAttr.dwBytesPerPixel);
	

	//Crear objecte d'imatge IPL
	IplImage* captura = cvCreateImage(cvSize(camAttr.dwWidth,camAttr.dwHeight), IPL_DEPTH_8U, 1);
	
	//Crear finestra
	cvNamedWindow("Camera",1);

	cout << "INSTRUCCIONS:" << endl;
	cout << "La finestra mostra una captura de la camera" << endl;
	cout << "Prem una tecla per mostrar el resultat de la binaritzacio" << endl;
	cout << "Torna a premer per realitzar una nova captura" << endl;
	cout << "Tanca aquesta finestra per sortir del proces" << endl;


	while(1){
		//Captura d'imatge al buffer
		m_cam->Snap(m_imgBuf);

		//calcular el llindar optim - Otsu		
		llindar = otsu(m_imgBuf, 1, camAttr.dwWidth * camAttr.dwHeight, 0, 0, camAttr.dwWidth * camAttr.dwHeight, 1, 0);

		//binaritzar	
		for(i=0; i<camAttr.dwWidth * camAttr.dwHeight; i++){		
			if((int)m_imgBuf[i] < llindar) //llindar
				m_imgBufBin[i]=0;
			else
				m_imgBufBin[i]=255;
		}

		
		//Volcar informaci capturada a la imatge IPL
		for(i=0; i < captura->imageSize; i++){
			captura->imageData[i] = m_imgBuf[i];
		}

		//Mostrar imatge original
		cvShowImage("Camera", captura);
		cvWaitKey(0);
		
		//Volcar informaci capturada i binaritzada a la imatge IPL
		for(i=0; i < captura->imageSize; i++){
			captura->imageData[i] = m_imgBufBin[i];
		}

		//Mostrar imatge binaritzada
		cvShowImage("Camera", captura);
		cvWaitKey(0);
	}
	
	return 0;	
}


/******************************************************
 * otsu
 *
 * Algoritme per trobar el llindar ptim per binaritzar la imatge.
 ******************************************************/
int otsu (unsigned char *image, int rows, int cols, int x0, int y0, int dx, int dy, int vvv)
{
	unsigned char *np;    // pointer to position in the image we are working with
	unsigned char op1, op2;   // predecessor of pixel *np (start value)
	int maxc=0;           // maximum contrast (start value) 
	int thresholdValue=1; // value we will threshold at
	int ihist[256];       // image histogram
	int chist[256];       // contrast histogram

	int i, j, k;          // various counters
	int is, i1, i2, ns, n1, n2, gmin, gmax;
	double m1, m2, sum, csum, fmax, sb;

	// zero out histogram ...
	memset(ihist, 0, sizeof(ihist));
	memset(chist, 0, sizeof(chist));
	op1=op2=0;

	gmin=255; gmax=0; k=1;//k=dy/512+1;
	// v0.43 first get max contrast, dont do it together with next step
	//  because it failes if we have pattern as background (on top)
	for (i =  0; i <  dy ; i+=k) {
	np = &image[(y0+i)*cols+x0];
	for (j = 0; j < dx ; j++) {
	  ihist[*np]++;
	  if(*np > gmax) gmax=*np;
	  if(*np < gmin) gmin=*np;
	  if (Abs(*np-op1)>maxc) maxc=Abs(*np-op1); /* new maximum contrast */
	  if (Abs(*np-op2)>maxc) maxc=Abs(*np-op2); /* new maximum contrast */
	  /* we hope that maxc will be find its maximum very fast */
	  op2=op1;    /* shift old pixel to next older */
	  op1=*np;     /* store old pixel for contrast check */
	  np++;       /* next pixel */
	}
	}

	// generate the histogram
	// Aug06 images with large white or black homogeneous
	//   areas give bad results, so we only add pixels on contrast edges
	for (i =  0; i <  dy ; i+=k) {
	np = &image[(y0+i)*cols+x0];
	for (j = 0; j < dx ; j++) {
	  if (Abs(*np-op1)>maxc/4
	   || Abs(*np-op2)>maxc/4)
		 chist[*np]++; // count only relevant pixels
	  op2=op1;    /* shift old pixel to next older */
	  op1=*np;     /* store old pixel for contrast check */
	  np++;       /* next pixel */
	}
	}

	// set up everything
	sum = csum = 0.0;
	ns = 0;
	is = 0;

	for (k = 0; k <= 255; k++) {
	sum += (double) k * (double) chist[k];  /* x*f(x) cmass moment */
	ns  += chist[k];                        /*  f(x)    cmass      */
	is  += ihist[k];                        /*  f(x)    imass      */
	// Debug: output to out_hist.dat?
	// fprintf(stderr,"\chistogram %3d %6d (brightness weight)", k, ihist[k]);
	}

	if (!ns) {
	// if n has no value we have problems...
	fprintf (stderr, "NOT NORMAL, thresholdValue = 160\n");
	return (160);
	}

	// ToDo: only care about extremas in a 3 pixel environment
	//       check if there are more than 2 mass centers (more colors)
	//       return object colors and color radius instead of threshold value
	//        also the reagion, where colored objects are found
	//       what if more than one background color? no otsu at all?
	//       whats background? box with lot of other boxes in it
	//       threshold each box (examples/invers.png,colors.png) 
	//  get maximum white and minimum black pixel color (possible range)
	//    check range between them for low..high contrast ???
	// typical scenes (which must be covered): 
	//    - white page with text of different colors (gray values)
	//    - binear page: background (gray=1) + black text (gray=0)
	//    - text mixed with big (dark) images
	//  ToDo: recursive clustering for maximum multipol moments?
	//  idea: normalize ihist to max=1024 before otsu?

	// do the otsu global thresholding method

	if ((vvv&1)) // Debug
	 fprintf(stderr,"# OTSU: value ihist chist mass_dipol_moment\n");
	fmax = -1.0;
	n1 = 0;
	for (k = 0; k < 255; k++) {
	n1 += chist[k];          // left  mass (integration)
	if (!n1) continue;       // we need at least one foreground pixel
	n2 = ns - n1;            // right mass (num pixels - left mass)
	if (n2 == 0) break;      // we need at least one background pixel
	csum += (double) k *chist[k];  // left mass moment
	m1 =        csum  / n1;        // left  mass center (black chars)
	m2 = (sum - csum) / n2;        // right mass center (white background)
	// max. dipol moment?
	// orig: sb = (double) n1 *(double) n2 * (m1 - m2) * (m1 - m2);
	sb = (double) n1 *(double) n2 * (m2 - m1); // seems to be better Aug06
	/* bbg: note: can be optimized. */
	if (sb > fmax) {
	  fmax = sb;
	  thresholdValue = k + 1;
	  // thresholdValue = (m1 + 3 * m2) / 4;
	}
	if ((vvv&1) && ihist[k]) // Debug
	 fprintf(stderr,"# OTSU: %3d %6d %6d %8.2f\n",
	   k, ihist[k], chist[k],
	   sb/(dx*dy));  /* normalized dipol moment */
	}
	// ToDo: error = left/right point where sb is 90% of maximum?
	// now we count all pixels for background detection
	i1 = 0;
	for (k = 0; k < thresholdValue; k++) {
	i1 += ihist[k];          // left  mass (integration)
	}
	i2 = is - i1;            // right mass (num pixels - left mass)

	// at this point we have our thresholding value
	// black_char: value<cs,  white_background: value>=cs

	// can it happen? check for sureness
	if (thresholdValue >  gmax) {
	fprintf(stderr,"# OTSU: thresholdValue >gmax\n");
	thresholdValue = gmax;
	}
	if (thresholdValue <= gmin) {
	fprintf(stderr,"# OTSU: thresholdValue<=gmin\n");
	thresholdValue = gmin+1;
	}

	// debug code to display thresholding values
	if ( vvv & 1 )
	fprintf(stderr,"# OTSU: thresholdValue = %d gmin=%d gmax=%d cmax=%d"
				 " i= %d %d\n",
	 thresholdValue, gmin, gmax, maxc, i1, i2);

	if (i1>=4*i2) { // black>=4*white, obviously black is background
	if ( vvv & 1 )
	  fprintf(stderr,"# OTSU: invert the image\n");
	// we do inversion here (no data lost)
	for (i =  0; i <  dy ; i++) {
	  np = &image[(y0+i)*cols+x0];
	  for (j = 0; j < dx ; j++) {
		*np=255-*np;
		np++;       /* next pixel */
	  }
	}
	thresholdValue=255-thresholdValue+1;
	}

	return(thresholdValue); 
	/* range: 0 < thresholdValue <= 255, example: 1 on b/w images */
	/* 0..threshold-1 is foreground */
	/* threshold..255 is background */
	/* ToDo:  min=blackmasscenter/2,thresh,max=(whitemasscenter+255)/2 */
}
